"""
This file defines variables for the modelled legislation.

A variable is a property of an Entity such as a 人物, a 世帯…

See https://openfisca.org/doc/key-concepts/variables.html
"""

# from cProfile import label
# from xmlrpc.client import Boolean
from datetime import date

# Import from numpy the operations you need to apply on OpenFisca"s population vectors
# Import from openfisca-core the Python objects used to code the legislation in OpenFisca
import numpy as np
from numpy import where
from openfisca_core.periods import DAY, ETERNITY
from openfisca_core.variables import Variable
# Import the Entities specifically defined for this tax and benefit system
from openfisca_japan.entities import 世帯, 人物


# This variable is a pure input: it doesn"t have a formula
class 誕生年月日(Variable):
    value_type = date
    default_value = date(1970, 1, 1)  # By default, if no value is set for a simulation, we consider the people involved in a simulation to be born on the 1st of Jan 1970.
    entity = 人物
    label = "人物の誕生年月日"
    definition_period = ETERNITY  # This variable cannot change over time.
    reference = "https://en.wiktionary.org/wiki/birthdate"


class 年齢(Variable):
    value_type = int
    entity = 人物
    definition_period = DAY
    label = "人物の年齢"

    def formula(対象人物, 対象期間, _parameters):
        誕生年月日 = 対象人物("誕生年月日", 対象期間)
        誕生年 = 誕生年月日.astype("datetime64[Y]").astype(int) + 1970
        誕生月 = 誕生年月日.astype("datetime64[M]").astype(int) % 12 + 1
        誕生日 = (誕生年月日 - 誕生年月日.astype("datetime64[M]") + 1).astype(int)

        誕生日を過ぎている = (誕生月 < 対象期間.start.month) + (誕生月 == 対象期間.start.month) * (誕生日 <= 対象期間.start.day)

        年齢 = (対象期間.start.year - 誕生年) - where(誕生日を過ぎている, 0, 1)  # If the birthday is not passed this year, subtract one year

        # NOTE: 誕生日が未来であった場合便宜上0歳として扱う(誤った情報が指定された場合でもOpenFiscaがクラッシュするのを防ぐため)
        return np.clip(年齢, 0, None)


# 小学n年生はn, 中学m年生はm+6, 高校l年生はl+9,
# 小学生未満は0以下の整数, 高校3年生より大きい学年は13以上の整数を返す
class 学年(Variable):
    value_type = int
    entity = 人物
    definition_period = DAY
    label = "人物の学年"

    def formula(対象人物, 対象期間, _parameters):
        誕生年月日 = 対象人物("誕生年月日", 対象期間)

        誕生年 = 誕生年月日.astype("datetime64[Y]").astype(int) + 1970
        誕生月 = 誕生年月日.astype("datetime64[M]").astype(int) % 12 + 1
        誕生日 = (誕生年月日 - 誕生年月日.astype("datetime64[M]") + 1).astype(int)

        # 早生まれは誕生月日が1/1~4/1
        早生まれ = (誕生月 < 4) + ((誕生月 == 4) * (誕生日 == 1))
        対象期間が四月以降 = 対象期間.start.month >= 4
        繰り上げ年数 = where(早生まれ, 1, 0) + where(対象期間が四月以降, 1, 0)

        return (対象期間.start.year - 誕生年) + 繰り上げ年数 - 7


class 世帯人数(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "世帯人数"

    def formula(対象世帯, 対象期間, parameters):
        # NOTE: OpenFiscaの以下のメソッドで直接計算可能だが、テストのinputで人数を直接指定できるようVariable化
        return 対象世帯.nb_persons()



"""
脳性まひ_進行性筋萎縮症の実装
"""

from openfisca_core.indexed_enums import Enum
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 人物


class 脳性まひ_進行性筋萎縮症パターン(Enum):
    __order__ = "無 有"
    無 = "無"
    有 = "有"


class 脳性まひ_進行性筋萎縮症(Variable):
    value_type = Enum
    possible_values = 脳性まひ_進行性筋萎縮症パターン
    default_value = 脳性まひ_進行性筋萎縮症パターン.無
    entity = 人物
    definition_period = DAY
    label = "人物の脳性まひ・進行性筋萎縮症パターン"



"""
児童手当の実装
"""

from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯
from openfisca_japan.variables.全般 import 高校生学年


class 児童手当(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "保護者への児童手当"
    reference = "https://www.cfa.go.jp/policies/kokoseido/jidouteate/annai/"
    documentation = """
    児童手当制度
    """

    def formula(対象世帯, 対象期間, parameters):
        児童手当 = parameters(対象期間).福祉.育児.児童手当

        年齢 = 対象世帯.members("年齢", 対象期間)
        学年 = 対象世帯.members("学年", 対象期間)

        # 児童手当金額
        二十二歳以下である = 対象世帯.has_role(世帯.子) * (年齢 <= 22)
        高校生以下である = (学年 <= 高校生学年.三年生.value) * (年齢 <= 18)
        # 二十二歳以下でない場合は -1 が入る
        二十二歳以下の出生順 = 対象世帯.get_rank(対象世帯, - 年齢, condition=二十二歳以下である)

        第二子以前である = (二十二歳以下の出生順 >= 0) * (二十二歳以下の出生順 < 2)
        高校生以下かつ第三子以降である = (二十二歳以下の出生順 >= 2) * 高校生以下である

        三歳未満かつ第二子以前である = (年齢 < 3) * 第二子以前である
        三歳以上かつ高校生以下かつ第二子以前である = (年齢 >= 3) * 高校生以下である * 第二子以前である
        児童手当金額 = 対象世帯.sum(高校生以下かつ第三子以降である * 児童手当.金額.高校生以下かつ第三子以降
                          + 三歳未満かつ第二子以前である * 児童手当.金額.三歳未満かつ第二子以前
                          + 三歳以上かつ高校生以下かつ第二子以前である * 児童手当.金額.三歳以上かつ高校生以下かつ第二子以前)

        return 児童手当金額



"""
先天性血液凝固因子障害等治療研究事業の実装
"""

from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯, 人物


class 先天性血液凝固因子障害等治療研究事業の対象者がいる(Variable):
    value_type = bool
    entity = 世帯
    definition_period = DAY
    label = "先天性血液凝固因子障害等治療研究事業の対象者の対象者がいるか否か"
    reference = "https://www.mhlw.go.jp/web/t_doc?dataId=00tb3840&dataType=1&pageNo=1"

    def formula(対象世帯, 対象期間, _parameters):
        先天性血液凝固因子障害等治療研究事業の対象者である = 対象世帯.members("先天性血液凝固因子障害等治療研究事業の対象者である", 対象期間)
        return 対象世帯.sum(先天性血液凝固因子障害等治療研究事業の対象者である)


class 先天性血液凝固因子障害等治療研究事業の対象者である(Variable):
    value_type = bool
    entity = 人物
    definition_period = DAY
    label = "先天性血液凝固因子障害等治療研究事業の対象者である"
    reference = "https://www.mhlw.go.jp/web/t_doc?dataId=00tb3840&dataType=1&pageNo=1"
    documentation = """
    医療費負担を軽減する制度であり他制度の計算結果と合計できないため、対象者であるかどうかのみ計算

    条件は以下リンクも参考になる。
    https://www.pref.aichi.jp/soshiki/kenkotaisaku/senketsu.html
    """

    def formula(対象人物, 対象期間, _parameters):
        先天性の血液凝固因子異常症である可能性がある = 対象人物("先天性の血液凝固因子異常症である可能性がある", 対象期間)
        年齢 = 対象人物("年齢", 対象期間)
        血液凝固因子異常症関連条件 = (年齢 >= 20) * 先天性の血液凝固因子異常症である可能性がある

        HIV感染者である = 対象人物("HIV感染者である", 対象期間)
        血液製剤の投与によってHIVに感染した = 対象人物("血液製剤の投与によってHIVに感染した", 対象期間)
        HIV関連条件 = HIV感染者である * 血液製剤の投与によってHIVに感染した

        return 血液凝固因子異常症関連条件 + HIV関連条件



"""
This file defines the entities needed by our legislation.

税金 and benefits can be calculated for different entities: 世帯, 人物, companies, etc.

See https://openfisca.org/doc/key-concepts/人物,_entities,_role.html
"""

from openfisca_core.entities import build_entity

世帯 = build_entity(
    key = "世帯",
    plural = "世帯一覧",
    label = "All the people in a family or group who live together in the same place.",
    doc = """
    世帯 is an example of a group entity.
    A group entity contains one or more individual·s.
    Each individual in a group entity has a role (e.g. parent or children). Some roles can only be held by a limited number of individuals (e.g. a "first_parent" can only be held by one individual), while others can have an unlimited number of individuals (e.g. "children").

    Example:
    Housing variables (e.g. 固定資産税") are usually defined for a group entity such as "世帯".

    Usage:
    Check the number of individuals of a specific role (e.g. check if there is a "second_parent" with 世帯.nb_人物(世帯.SECOND_PARENT)).
    Calculate a variable applied to each individual of the group entity (e.g. calculate the "所得" of each member of the "世帯" with salaries = 世帯.members("所得", period = MONTH); sum_salaries = 世帯.sum(salaries)).

    For more information, see: https://openfisca.org/doc/coding-the-legislation/50_entities.html
    """,
    roles = [
        {
            "key": "親",  # 子供がいない場合も大人は「親」entityとする
            "plural": "親一覧",
            "label": "親",
            "max": 2,
            "doc": "大人あるいは子を持つ親",
            },
        {
            "key": "子",
            "plural": "子一覧",
            "label": "子",
            "doc": "親と同居している子供",
            },
        {
            "key": "祖父母",
            "plural": "祖父母一覧",
            "label": "祖父母",
            "doc": "Other individuals living in the 世帯.",
            },
        ],
    )

人物 = build_entity(
    key = "人物",
    plural = "世帯員",
    label = "An individual. The minimal legal entity on which a legislation might be applied.",
    doc = """

    Variables like "所得" and "所得税" are usually defined for the entity "人物".

    Usage:
    Calculate a variable applied to a "人物" (e.g. access the "所得" of a specific month with 人物("所得", "2017-05")).
    Check the role of a "人物" in a group entity (e.g. check if a the "人物" is a "first_parent" in a "世帯" entity with 人物.has_role(世帯.FIRST_PARENT)).

    For more information, see: https://openfisca.org/doc/coding-the-legislation/50_entities.html
    """,
    is_person = True,
    )

entities = [世帯, 人物]



"""
障害児福祉手当の実装
"""
import numpy as np
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯, 人物
from openfisca_japan.variables.障害.愛の手帳 import 愛の手帳等級パターン
from openfisca_japan.variables.障害.療育手帳 import 療育手帳等級パターン
from openfisca_japan.variables.障害.身体障害者手帳 import 身体障害者手帳等級パターン


class 障害児福祉手当(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "障害児童福祉手当"
    reference = "https://www.mhlw.go.jp/bunya/shougaihoken/jidou/hukushi.html"
    documentation = """
    手帳について要件は無いが、目安となる等級をもとに算出
    https://h-navi.jp/column/article/35029230
    """

    # TODO: 重度心身障害者手当所得制限と同じ控除を適用する
    # https://www.fukushi.metro.tokyo.lg.jp/shinsho/teate/jidou.html

    def formula(対象世帯, 対象期間, parameters):
        障害児福祉手当 = parameters(対象期間).福祉.育児.障害児福祉手当

        所得一覧 = 対象世帯.members("所得", 対象期間)
        身体障害者手帳等級一覧 = 対象世帯.members("身体障害者手帳等級", 対象期間)
        療育手帳等級一覧 = 対象世帯.members("療育手帳等級", 対象期間)
        愛の手帳等級一覧 = 対象世帯.members("愛の手帳等級", 対象期間)
        年齢 = 対象世帯.members("年齢", 対象期間)
        上限年齢未満 = 年齢 < 障害児福祉手当.上限年齢

        対象障害者手帳等級 = \
            (身体障害者手帳等級一覧 == 身体障害者手帳等級パターン.一級) + \
            (身体障害者手帳等級一覧 == 身体障害者手帳等級パターン.二級) + \
            (療育手帳等級一覧 == 療育手帳等級パターン.A) + \
            (療育手帳等級一覧 == 療育手帳等級パターン.B) + \
            (愛の手帳等級一覧 == 愛の手帳等級パターン.一度) + \
            (愛の手帳等級一覧 == 愛の手帳等級パターン.二度)

        対象障害者である = 上限年齢未満 * 対象障害者手帳等級
        # 障害児はほとんど控除が該当しないと考える
        対象障害者の所得 = 対象障害者である * 所得一覧

        受給者の所得制限限度額 = 対象世帯.members("障害児福祉手当_受給者の所得制限限度額", 対象期間)
        受給者の所得条件 = 対象障害者の所得 < 受給者の所得制限限度額

        # 世帯高所得は保護者によるものだと想定する
        世帯高所得 = 対象世帯("控除後世帯高所得", 対象期間)
        扶養義務者の所得制限限度額 = 対象世帯("障害児福祉手当_扶養義務者の所得制限限度額", 対象期間)
        扶養義務者の所得条件 = 世帯高所得 < 扶養義務者の所得制限限度額

        # 障害児福祉手当は対象障害者が受給者のため、世帯では対象障害者の人数分支給される。
        # NOTE: 受給者の所得条件は人物、扶養義務者の所得条件は世帯ごとに計算しているため式中の登場個所が異なる
        手当金額 = 対象世帯.sum(受給者の所得条件 * 対象障害者である * 障害児福祉手当.金額)

        return 扶養義務者の所得条件 * 手当金額


class 障害児福祉手当_受給者の所得制限限度額(Variable):
    value_type = float
    entity = 人物
    definition_period = DAY
    label = "障害児童福祉手当における、受給者の所得制限限度額"
    reference = "https://www.mhlw.go.jp/bunya/shougaihoken/jidou/hukushi.html"

    def formula(対象人物, 対象期間, parameters):
        扶養人数 = 対象人物.世帯("扶養人数", 対象期間)

        # NOTE: 直接 `受給者[扶養人数]` のように要素参照すると型が合わず複数世帯の場合に計算できないためnp.selectを使用
        障害児福祉手当 = parameters(対象期間).福祉.育児.障害児福祉手当
        受給者の所得制限限度額 = np.select(
            [扶養人数 == i for i in range(10)],
            [障害児福祉手当.所得制限限度額.受給者[i] for i in range(10)],
            -1).astype(int)

        return 受給者の所得制限限度額


class 障害児福祉手当_扶養義務者の所得制限限度額(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "障害児童福祉手当における、扶養義務者の所得制限限度額"
    reference = "https://www.mhlw.go.jp/bunya/shougaihoken/jidou/hukushi.html"
    documentation = """
    便宜上最も所得が多い世帯員を扶養義務者としているため、この値は世帯ごとに計算される
    """

    def formula(対象世帯, 対象期間, parameters):
        扶養人数 = 対象世帯("扶養人数", 対象期間)

        # NOTE: 直接 `受給者[扶養人数]` のように要素参照すると型が合わず複数世帯の場合に計算できないためnp.selectを使用
        障害児福祉手当 = parameters(対象期間).福祉.育児.障害児福祉手当
        扶養義務者の所得制限限度額 = np.select(
            [扶養人数 == i for i in range(10)],
            [障害児福祉手当.所得制限限度額.扶養義務者[i] for i in range(10)],
            -1).astype(int)

        return 扶養義務者の所得制限限度額



"""
重度心身障害者手当の実装
"""

import numpy as np
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯, 人物
from openfisca_japan.variables.障害.愛の手帳 import 愛の手帳等級パターン
from openfisca_japan.variables.障害.身体障害者手帳 import 身体障害者手帳等級パターン


class 重度心身障害者手当_最大(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "重度心身障害者手当の最大額"
    reference = "https://www.fukushi.metro.tokyo.lg.jp/shinsho/teate/juudo.html"
    documentation = """
    東京都の制度
    厳密な判定には詳細な症状が必要なため、愛の手帳等級、身体障害者手帳等から推定可能な最小値、最大値を算出

    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/file/06-Seisakujouhou-12200000-Shakaiengokyokushougaihokenfukushibu/0000172197.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        居住都道府県 = 対象世帯("居住都道府県", 対象期間)
        # 東京都以外は対象外
        居住地条件 = 居住都道府県 == "東京都"

        年齢 = 対象世帯.members("年齢", 対象期間)
        年齢条件 = 年齢 < 65

        愛の手帳等級 = 対象世帯.members("愛の手帳等級", 対象期間)
        身体障害者手帳等級 = 対象世帯.members("身体障害者手帳等級", 対象期間)
        # 該当する可能性のある条件
        一号要件 = (愛の手帳等級 == 愛の手帳等級パターン.一度) | (愛の手帳等級 == 愛の手帳等級パターン.二度)
        二号要件 = ((愛の手帳等級 == 愛の手帳等級パターン.一度) | (愛の手帳等級 == 愛の手帳等級パターン.二度)) * \
            ((身体障害者手帳等級 == 身体障害者手帳等級パターン.一級) | (身体障害者手帳等級 == 身体障害者手帳等級パターン.二級))
        三号要件 = 身体障害者手帳等級 == 身体障害者手帳等級パターン.一級
        障害条件 = 一号要件 | 二号要件 | 三号要件

        所得条件 = 対象世帯.members("重度心身障害者手当所得制限", 対象期間)

        人物ごとの受給条件 = 年齢条件 * 障害条件 * 所得条件
        # NOTE: 居住地条件は世帯の条件のため、人物ごとの受給条件とは型が合わず直接計算できない
        対象人数 = 対象世帯.sum(人物ごとの受給条件) * 居住地条件

        return 対象人数 * parameters(対象期間).東京都.福祉.重度心身障害者手当.重度心身障害者手当額


class 重度心身障害者手当_最小(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "重度心身障害者手当の最小額"
    reference = "https://www.fukushi.metro.tokyo.lg.jp/shinsho/teate/juudo.html"
    documentation = """
    東京都の制度
    厳密な判定には詳細な症状が必要なため、愛の手帳等級、身体障害者手帳等から推定可能な最小値、最大値を算出

    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/file/06-Seisakujouhou-12200000-Shakaiengokyokushougaihokenfukushibu/0000172197.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        居住都道府県 = 対象世帯("居住都道府県", 対象期間)
        # 東京都以外は対象外
        # 東京都以外は対象外
        居住地条件 = 居住都道府県 == "東京都"

        年齢 = 対象世帯.members("年齢", 対象期間)
        年齢条件 = 年齢 < 65

        愛の手帳等級 = 対象世帯.members("愛の手帳等級", 対象期間)
        身体障害者手帳等級 = 対象世帯.members("身体障害者手帳等級", 対象期間)
        # 1号要件,3号要件は愛の手帳、身体障害者手帳のみでは該当しない可能性があるため最小額は0
        二号要件 = ((愛の手帳等級 == 愛の手帳等級パターン.一度) | (愛の手帳等級 == 愛の手帳等級パターン.二度)) * \
            ((身体障害者手帳等級 == 身体障害者手帳等級パターン.一級) | (身体障害者手帳等級 == 身体障害者手帳等級パターン.二級))
        障害条件 = 二号要件

        所得条件 = 対象世帯.members("重度心身障害者手当所得制限", 対象期間)

        人物ごとの受給条件 = 年齢条件 * 障害条件 * 所得条件
        # NOTE: 居住地条件は世帯の条件のため、人物ごとの受給条件とは型が合わず直接計算できない
        対象人数 = 対象世帯.sum(人物ごとの受給条件) * 居住地条件

        return 対象人数 * parameters(対象期間).東京都.福祉.重度心身障害者手当.重度心身障害者手当額


class 重度心身障害者手当所得制限(Variable):
    value_type = bool
    entity = 人物
    definition_period = DAY
    label = "重度心身障害者手当の所得制限"
    reference = "https://www.fukushi.metro.tokyo.lg.jp/shinsho/teate/juudo.html"

    # TODO: 障害児福祉手当も同じ控除を適用する。
    # https://www.fukushi.metro.tokyo.lg.jp/shinsho/teate/jidou.html
    # 心身障害者手当も同じ控除になるか要調査

    def formula(対象人物, 対象期間, parameters):
        重度心身障害者手当 = parameters(対象期間).東京都.福祉.重度心身障害者手当
        扶養人数 = 対象人物.世帯("扶養人数", 対象期間)
        # 複数世帯入力(2以上の長さのndarray入力)対応のためndarray化
        所得制限限度額 = np.array(重度心身障害者手当.所得制限限度額)[np.clip(扶養人数, 0, 5)]

        所得 = 対象人物("所得", 対象期間)
        # 便宜上、世帯の最大の所得を扶養義務者の所得とする
        扶養義務者の所得 = 対象人物.世帯("世帯高所得", 対象期間)
        年齢 = 対象人物("年齢", 対象期間)
        # 20歳未満の場合扶養義務者、20歳以上の場合本人の所得を参照
        本人所得を参照 = 年齢 >= 20
        対象となる所得 = np.where(本人所得を参照, 所得, 扶養義務者の所得)

        給与所得及び雑所得からの控除額 = parameters(対象期間).所得.給与所得及び雑所得からの控除額
        社会保険料控除 = np.where(np.logical_not(本人所得を参照), parameters(対象期間).所得.社会保険料相当額, 0)
        # 上限33万円
        配偶者特別控除 = np.clip(対象人物("配偶者特別控除", 対象期間), 0, 330000)

        障害者控除 = 対象人物("障害者控除", 対象期間)

        # 障害者控除, 特別障害者控除については、本人所得の場合本人の分は適用しない
        障害者控除対象 = 対象人物("障害者控除対象", 対象期間)
        特別障害者控除対象 = 対象人物("特別障害者控除対象", 対象期間)
        本人の控除 = np.select(
            [障害者控除対象, 特別障害者控除対象],
            [parameters(対象期間).所得.障害者控除額, parameters(対象期間).所得.特別障害者控除額],
            0)
        障害者控除 = 障害者控除 - 本人所得を参照 * 本人の控除

        # TODO: 同居特別障害者の場合も通常の特別障害者と同じ額に補正する必要がある？
        # 単純に記載を省略しているだけの可能性もあり (https://www.fukushi.metro.tokyo.lg.jp/shinsho/teate/toku_shou.html で国の制度を照会しているが同居特別障害者の記載はない)

        寡婦控除 = 対象人物("寡婦控除", 対象期間)
        ひとり親控除 = 対象人物("ひとり親控除", 対象期間)
        勤労学生控除 = 対象人物("勤労学生控除", 対象期間)

        # 他の控除（雑損控除・医療費控除等）は定額でなく実費を元に算出するため除外する
        控除総額 = 給与所得及び雑所得からの控除額 + 社会保険料控除 + 配偶者特別控除 + 障害者控除 + 寡婦控除 + ひとり親控除 + 勤労学生控除
        控除後所得 = 対象となる所得 - 控除総額

        return 控除後所得 <= 所得制限限度額



"""
内部障害の実装
"""

from openfisca_core.indexed_enums import Enum
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 人物


class 内部障害パターン(Enum):
    __order__ = "無 有"
    無 = "無"
    有 = "有"


class 内部障害(Variable):
    value_type = Enum
    possible_values = 内部障害パターン
    default_value = 内部障害パターン.無
    entity = 人物
    definition_period = DAY
    label = "人物の内部障害パターン"



"""
被災者の住宅に関するVariableを定義
"""

from openfisca_core.indexed_enums import Enum
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯, 人物


class 住宅被害パターン(Enum):
    __order__ = "無 中規模半壊 大規模半壊 長期避難 解体 全壊 滅失または流失"
    無 = "無"
    中規模半壊 = "中規模半壊"
    大規模半壊 = "大規模半壊"
    長期避難 = "長期避難"
    解体 = "解体"
    全壊 = "全壊"
    滅失または流失 = "滅失または流失"


class 住宅被害(Variable):
    value_type = Enum
    possible_values = 住宅被害パターン
    default_value = 住宅被害パターン.無
    entity = 世帯
    definition_period = DAY
    label = "被災者の住宅被害"


class 住宅再建方法パターン(Enum):
    __order__ = "無 建設または購入 補修 賃借"
    無 = "無"
    建設または購入 = "建設または購入"
    補修 = "補修"
    賃借 = "賃借"


class 住宅再建方法(Variable):
    value_type = Enum
    possible_values = 住宅再建方法パターン
    default_value = 住宅再建方法パターン.無
    entity = 世帯
    definition_period = DAY
    label = "被災者の住宅再建方法"


class 家財の損害パターン(Enum):
    # TODO: 他制度でより細かい分類が必要になったら追加
    __order__ = "無 三分の一未満 三分の一以上"
    無 = "無"
    三分の一未満 = "三分の一未満"
    三分の一以上 = "三分の一以上"


class 家財の損害(Variable):
    value_type = Enum
    possible_values = 家財の損害パターン
    default_value = 家財の損害パターン.無
    entity = 世帯
    definition_period = DAY
    label = "被災者の住宅の家財の損害"


class 災害による負傷の療養期間パターン(Enum):
    # TODO: 他制度でより細かい分類が必要になったら追加
    __order__ = "無 一か月未満 一か月以上"
    無 = "無"
    一か月未満 = "一か月未満"
    一か月以上 = "一か月以上"


class 災害による負傷の療養期間(Variable):
    value_type = Enum
    possible_values = 災害による負傷の療養期間パターン
    default_value = 災害による負傷の療養期間パターン.無
    entity = 人物
    definition_period = DAY
    label = "災害により負傷した被災者の療養期間"



"""
This file defines our country's tax and benefit system.

A tax and benefit system is the higher-level instance in OpenFisca.
Its goal is to model the legislation of a country.
Basically a tax and benefit system contains simulation variables (source code) and legislation parameters (data).

See https://openfisca.org/doc/key-concepts/tax_and_benefit_system.html
"""

import os

from openfisca_core.taxbenefitsystems import TaxBenefitSystem
from openfisca_japan import entities
from openfisca_japan.situation_examples import 夫婦

COUNTRY_DIR = os.path.dirname(os.path.abspath(__file__))


# Our country tax and benefit class inherits from the general TaxBenefitSystem class.
# The name CountryTaxBenefitSystem must not be changed, as all tools of the OpenFisca ecosystem expect a CountryTaxBenefitSystem class to be exposed in the __init__ module of a country package.
class CountryTaxBenefitSystem(TaxBenefitSystem):
    def __init__(self):
        # We initialize our tax and benefit system with the general constructor
        super().__init__(entities.entities)

        # We add to our tax and benefit system all the variables
        self.add_variables_from_directory(os.path.join(COUNTRY_DIR, "variables"))

        # We add to our tax and benefit system all the legislation parameters defined in the  parameters files
        param_path = os.path.join(COUNTRY_DIR, "parameters")
        self.load_parameters(param_path)

        # We define which variable, parameter and simulation example will be used in the OpenAPI specification
        self.open_api_config = {
            "variable_example": "所得",
            "parameter_example": "所得.寡婦控除額",
            "simulation_example": 夫婦,
            }



"""
精神障害者保健福祉手帳の実装
"""

from openfisca_core.indexed_enums import Enum
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 人物


class 精神障害者保健福祉手帳等級パターン(Enum):
    __order__ = "無 一級 二級 三級"
    無 = "無"
    一級 = "一級"
    二級 = "二級"
    三級 = "三級"


class 精神障害者保健福祉手帳等級(Variable):
    value_type = Enum
    possible_values = 精神障害者保健福祉手帳等級パターン
    default_value = 精神障害者保健福祉手帳等級パターン.無
    entity = 人物
    definition_period = DAY
    label = "人物の精神障害者保健福祉手帳の等級"



"""
高等学校奨学給付金の実装
"""

from functools import cache

import numpy as np
from openfisca_core.indexed_enums import Enum
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan import COUNTRY_DIR
from openfisca_japan.entities import 世帯, 人物
from openfisca_japan.variables.全般 import 高校生学年


@cache
def 国立高等学校奨学給付金表():
    """
    csvファイルから値を読み込み

    国立高等学校奨学給付金表()[世帯区分, 履修形態] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/育児/高等学校奨学給付金/国立高等学校奨学給付金額.csv",
                         delimiter=",", skip_header=1, dtype="int64")[:, 1:]


@cache
def 公立高等学校奨学給付金表():
    """
    csvファイルから値を読み込み

    公立高等学校奨学給付金表()[世帯区分, 履修形態] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/育児/高等学校奨学給付金/公立高等学校奨学給付金額.csv",
                         delimiter=",", skip_header=1, dtype="int64")[:, 1:]


@cache
def 私立高等学校奨学給付金表():
    """
    csvファイルから値を読み込み

    私立高等学校奨学給付金表()[世帯区分, 履修形態] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/育児/高等学校奨学給付金/私立高等学校奨学給付金額.csv",
                         delimiter=",", skip_header=1, dtype="int64")[:, 1:]


class 高等学校奨学給付金_最小(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "高等学校奨学給付金"
    reference = "https://www.mext.go.jp/a_menu/shotou/mushouka/1344089.htm"
    documentation = """
    高等学校奨学給付金_最小
    (東京都HP)https://www.kyoiku.metro.tokyo.lg.jp/admission/tuition/tuition/scholarship_public_school.html
    (兵庫HP)https://web.pref.hyogo.lg.jp/kk35/shougakukyuuhukinn.html
    """

    def formula(対象世帯, 対象期間, _parameters):
        生活保護受給可能 = 対象世帯("生活保護", 対象期間) > 0
        生活保護受給世帯の高等学校奨学給付金 = \
            対象世帯.sum(対象世帯.members("生活保護受給世帯の高等学校奨学給付金", 対象期間)) * 生活保護受給可能

        住民税非課税世帯である = 対象世帯("住民税非課税世帯", 対象期間)
        住民税非課税世帯の高等学校奨学給付金 = \
            対象世帯.sum(対象世帯.members("住民税非課税世帯の高等学校奨学給付金", 対象期間)) * 住民税非課税世帯である

        # 金額が小さい方(ただし、小さい方の金額が0円の場合はもう片方を適用)
        年間支給金額 = np.select(
            [(生活保護受給世帯の高等学校奨学給付金 >= 住民税非課税世帯の高等学校奨学給付金) * 住民税非課税世帯である,
             (生活保護受給世帯の高等学校奨学給付金 < 住民税非課税世帯の高等学校奨学給付金) * 生活保護受給可能,
             住民税非課税世帯の高等学校奨学給付金 == 0,
             生活保護受給世帯の高等学校奨学給付金 == 0],
            [住民税非課税世帯の高等学校奨学給付金,
             生活保護受給世帯の高等学校奨学給付金,
             生活保護受給世帯の高等学校奨学給付金,
             住民税非課税世帯の高等学校奨学給付金],
            0)

        # 月間支給金額へ変換
        return np.floor(年間支給金額 / 12)


class 高等学校奨学給付金_最大(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "高等学校奨学給付金"
    reference = "https://www.mext.go.jp/a_menu/shotou/mushouka/1344089.htm"
    documentation = """
    高等学校奨学給付金_最大
    (東京都HP)https://www.kyoiku.metro.tokyo.lg.jp/admission/tuition/tuition/scholarship_public_school.html
    (兵庫HP)https://web.pref.hyogo.lg.jp/kk35/shougakukyuuhukinn.html
    """

    def formula(対象世帯, 対象期間, _parameters):
        生活保護受給可能 = 対象世帯("生活保護", 対象期間) > 0
        生活保護受給世帯の高等学校奨学給付金 = \
            対象世帯.sum(対象世帯.members("生活保護受給世帯の高等学校奨学給付金", 対象期間)) * 生活保護受給可能

        住民税非課税世帯である = 対象世帯("住民税非課税世帯", 対象期間)
        住民税非課税世帯の高等学校奨学給付金 = \
            対象世帯.sum(対象世帯.members("住民税非課税世帯の高等学校奨学給付金", 対象期間)) * 住民税非課税世帯である

        # 金額が大きい方
        年間支給金額 = np.select(
            [生活保護受給世帯の高等学校奨学給付金 >= 住民税非課税世帯の高等学校奨学給付金,
             生活保護受給世帯の高等学校奨学給付金 < 住民税非課税世帯の高等学校奨学給付金],
            [生活保護受給世帯の高等学校奨学給付金,
             住民税非課税世帯の高等学校奨学給付金],
            0)

        # 月間支給金額へ変換
        return np.floor(年間支給金額 / 12)


class 生活保護受給世帯の高等学校奨学給付金(Variable):
    value_type = int
    entity = 人物
    definition_period = DAY
    label = "生活保護受給世帯の高等学校奨学給付金"
    reference = "https://www.mext.go.jp/a_menu/shotou/mushouka/1344089.htm"
    documentation = """
    生活保護受給世帯の世帯員の高等学校奨学給付金
    (東京都HP)https://www.kyoiku.metro.tokyo.lg.jp/admission/tuition/tuition/scholarship_public_school.html
    (兵庫HP)https://web.pref.hyogo.lg.jp/kk35/shougakukyuuhukinn.html
    """

    def formula(対象人物, 対象期間, parameters):
        子供である = 対象人物.has_role(世帯.子)
        高校生である = 対象人物("高校生である", 対象期間)

        高校履修種別 = 対象人物("高校履修種別", 対象期間).decode()
        高校履修種別区分 = np.select(
            [高校履修種別 == 高校履修種別パターン.全日制課程,
             高校履修種別 == 高校履修種別パターン.定時制課程,
             高校履修種別 == 高校履修種別パターン.通信制課程,
             高校履修種別 == 高校履修種別パターン.専攻科],
            [0, 1, 2, 3],
            -1).astype(int)  # intにできるようデフォルトをNoneではなく-1

        高校運営種別 = 対象人物("高校運営種別", 対象期間).decode()

        支給対象世帯区分 = 0  # 生活保護世帯に対応

        年間給付金額 = np.select(
            [高校運営種別 == 高校運営種別パターン.国立,
             高校運営種別 == 高校運営種別パターン.公立,
             高校運営種別 == 高校運営種別パターン.私立],
            [国立高等学校奨学給付金表()[支給対象世帯区分, 高校履修種別区分],
             公立高等学校奨学給付金表()[支給対象世帯区分, 高校履修種別区分],
             私立高等学校奨学給付金表()[支給対象世帯区分, 高校履修種別区分]],
            0)

        return 年間給付金額 * 高校生である * 子供である


class 住民税非課税世帯の高等学校奨学給付金(Variable):
    value_type = int
    entity = 人物
    definition_period = DAY
    label = "住民税非課税世帯の高等学校奨学給付金"
    reference = "https://www.mext.go.jp/a_menu/shotou/mushouka/1344089.htm"
    documentation = """
    住民税非課税世帯の世帯員の高等学校奨学給付金
    (東京都HP)https://www.kyoiku.metro.tokyo.lg.jp/admission/tuition/tuition/scholarship_public_school.html
    (兵庫HP)https://web.pref.hyogo.lg.jp/kk35/shougakukyuuhukinn.html
    (東京都私立財団HP)https://www.shigaku-tokyo.or.jp/pa_shougaku.html
    """

    def formula(対象人物, 対象期間, parameters):
        子供である = 対象人物.has_role(世帯.子)
        年齢 = 対象人物("年齢", 対象期間)
        支給対象である = 子供である & (年齢 < 23)
        子供の年齢降順インデックス = 対象人物.get_rank(対象人物.世帯, -年齢, condition=支給対象である)
        高校生である = 対象人物("高校生である", 対象期間)

        高校履修種別 = 対象人物("高校履修種別", 対象期間).decode()
        高校履修種別区分 = np.select(
            [高校履修種別 == 高校履修種別パターン.全日制課程,
             高校履修種別 == 高校履修種別パターン.定時制課程,
             高校履修種別 == 高校履修種別パターン.通信制課程,
             高校履修種別 == 高校履修種別パターン.専攻科],
            [0, 1, 2, 3],
            -1).astype(int)  # intにできるようデフォルトをNoneではなく-1

        高校運営種別 = 対象人物("高校運営種別", 対象期間).decode()

        通信制課程の高校に通う世帯員がいる = 対象人物.世帯("通信制課程の高校に通う世帯員がいる", 対象期間)
        # NOTE: 支給対象の範囲内の第一子か否かで区分が変わる(インデックスは0始まりのため、0は第一子を意味する)
        支給対象の範囲内の第一子である = 子供の年齢降順インデックス == 0
        第一子扱いとなる高校履修種別である = (高校履修種別 == 高校履修種別パターン.全日制課程) + (高校履修種別 == 高校履修種別パターン.定時制課程)
        支給対象世帯区分 = np.select(
            [支給対象の範囲内の第一子である * np.logical_not(通信制課程の高校に通う世帯員がいる) * 第一子扱いとなる高校履修種別である],
            [1],
            2).astype(int)

        年間給付金額 = np.select(
            [高校運営種別 == 高校運営種別パターン.国立,
             高校運営種別 == 高校運営種別パターン.公立,
             高校運営種別 == 高校運営種別パターン.私立],
            [国立高等学校奨学給付金表()[支給対象世帯区分, 高校履修種別区分],
             公立高等学校奨学給付金表()[支給対象世帯区分, 高校履修種別区分],
             私立高等学校奨学給付金表()[支給対象世帯区分, 高校履修種別区分]],
            0)

        return 年間給付金額 * 高校生である * 子供である


class 高校履修種別パターン(Enum):
    __order__ = "無 全日制課程 定時制課程 通信制課程 専攻科"
    無 = "無"
    全日制課程 = "全日制課程"
    定時制課程 = "定時制課程"
    通信制課程 = "通信制課程"
    専攻科 = "専攻科"


class 高校履修種別(Variable):
    value_type = Enum
    possible_values = 高校履修種別パターン
    default_value = 高校履修種別パターン.無
    entity = 人物
    definition_period = DAY
    label = "高校履修種別"


class 高校運営種別パターン(Enum):
    __order__ = "無 国立 公立 私立"
    無 = "無"
    国立 = "国立"
    公立 = "公立"
    私立 = "私立"


class 高校運営種別(Variable):
    value_type = Enum
    possible_values = 高校運営種別パターン
    default_value = 高校運営種別パターン.無
    entity = 人物
    definition_period = DAY
    label = "高校運営種別"


class 支給対象世帯(Enum):
    __order__ = "生活保護世帯 非課税世帯1 非課税世帯2"
    生活保護世帯 = "生活保護（生業扶助）受給世帯"
    非課税世帯1 = "非課税世帯（第1子）"
    非課税世帯2 = "非課税世帯（第2子）"


class 高校生である(Variable):
    value_type = int
    entity = 人物
    definition_period = DAY
    label = "高校生であるかどうか"

    def formula(対象人物, 対象期間, _parameters):
        学年 = 対象人物("学年", 対象期間)
        高校履修種別 = 対象人物("高校履修種別", 対象期間)
        return (学年 >= 高校生学年.一年生.value) * (学年 <= 高校生学年.三年生.value) * (高校履修種別 != 高校履修種別パターン.無)


class 通信制課程の高校に通う世帯員がいる(Variable):
    value_type = bool
    entity = 世帯
    definition_period = DAY
    label = "通信制課程の高校に通う世帯員がいる"

    def formula(対象世帯, 対象期間, _parameters):
        高校履修種別一覧 = 対象世帯.members("高校履修種別", 対象期間).decode()
        return 対象世帯.any(高校履修種別一覧 == 高校履修種別パターン.通信制課程)



"""
腎不全の実装
"""

from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 人物


class 慢性腎不全である(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "慢性腎不全である"


class 人工透析を行っている(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "人工透析を行っている"



"""
児童扶養手当の実装
"""

import numpy as np
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯
from openfisca_japan.variables.障害.愛の手帳 import 愛の手帳等級パターン
from openfisca_japan.variables.障害.療育手帳 import 療育手帳等級パターン
from openfisca_japan.variables.障害.精神障害者保健福祉手帳 import 精神障害者保健福祉手帳等級パターン
from openfisca_japan.variables.障害.身体障害者手帳 import 身体障害者手帳等級パターン


class 児童扶養手当_最大(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "保護者への児童扶養手当の最大額"
    reference = "https://www.cfa.go.jp/policies/hitori-oya/fuyou-teate/"
    documentation = """
    児童扶養手当制度
    """

    def formula(対象世帯, 対象期間, parameters):
        児童扶養手当 = parameters(対象期間).福祉.育児.児童扶養手当

        全部支給所得条件 = 対象世帯("児童扶養手当の全部支給所得条件", 対象期間)
        一部支給所得条件 = 対象世帯("児童扶養手当の一部支給所得条件", 対象期間)
        ひとり親世帯である = 対象世帯("ひとり親", 対象期間)
        手当条件 = ひとり親世帯である * (全部支給所得条件 + 一部支給所得条件)

        最大支給額児童1人 = 全部支給所得条件 * 児童扶養手当.金額.全部支給.児童1人 + 一部支給所得条件 * 児童扶養手当.金額.一部支給_最大額.児童1人
        最大支給額児童2人 = 全部支給所得条件 * 児童扶養手当.金額.全部支給.児童2人 + 一部支給所得条件 * 児童扶養手当.金額.一部支給_最大額.児童2人
        最大支給額児童3人目以降 = 全部支給所得条件 * 児童扶養手当.金額.全部支給.児童3人目以降 + 一部支給所得条件 * 児童扶養手当.金額.一部支給_最大額.児童3人目以降

        対象児童人数 = 対象世帯("児童扶養手当の対象児童人数", 対象期間)

        # 児童の人数に応じて手当金額を変える
        # TODO: 一部支給の場合に対応する。(手当額の算出方法不明)
        # TODO: 公的年金額が児童扶養手当額より低い場合はその差額を児童扶養手当として受け取れる
        # 参考：https://www.city.shibuya.tokyo.jp/kodomo/teate/hitorioya/heikyu.html
        手当金額 = \
            ((対象児童人数 == 1) * 最大支給額児童1人) +\
            ((対象児童人数 == 2) * (最大支給額児童1人 + 最大支給額児童2人)) +\
            ((対象児童人数 > 2) * (最大支給額児童1人 + 最大支給額児童2人 + (最大支給額児童3人目以降 * (対象児童人数 - 2))))

        return 手当条件 * 手当金額


class 児童扶養手当_最小(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "保護者への児童扶養手当の最小額"
    reference = "https://www.cfa.go.jp/policies/hitori-oya/fuyou-teate/"
    documentation = """
    児童扶養手当制度
    """

    def formula(対象世帯, 対象期間, parameters):
        児童扶養手当 = parameters(対象期間).福祉.育児.児童扶養手当

        全部支給所得条件 = 対象世帯("児童扶養手当の全部支給所得条件", 対象期間)
        一部支給所得条件 = 対象世帯("児童扶養手当の一部支給所得条件", 対象期間)
        ひとり親世帯である = 対象世帯("ひとり親", 対象期間)
        手当条件 = ひとり親世帯である * (全部支給所得条件 + 一部支給所得条件)

        最小支給額児童1人 = 全部支給所得条件 * 児童扶養手当.金額.全部支給.児童1人 + 一部支給所得条件 * 児童扶養手当.金額.一部支給_最小額.児童1人
        最小支給額児童2人 = 全部支給所得条件 * 児童扶養手当.金額.全部支給.児童2人 + 一部支給所得条件 * 児童扶養手当.金額.一部支給_最小額.児童2人
        最小支給額児童3人目以降 = 全部支給所得条件 * 児童扶養手当.金額.全部支給.児童3人目以降 + 一部支給所得条件 * 児童扶養手当.金額.一部支給_最小額.児童3人目以降

        対象児童人数 = 対象世帯("児童扶養手当の対象児童人数", 対象期間)

        # 児童の人数に応じて手当金額を変える
        # TODO: 一部支給の場合に対応する。(手当額の算出方法不明)
        # TODO: 公的年金額が児童扶養手当額より低い場合はその差額を児童扶養手当として受け取れる
        # 参考：https://www.city.shibuya.tokyo.jp/kodomo/teate/hitorioya/heikyu.html
        手当金額 = \
            ((対象児童人数 == 1) * 最小支給額児童1人) +\
            ((対象児童人数 == 2) * (最小支給額児童1人 + 最小支給額児童2人)) +\
            ((対象児童人数 > 2) * (最小支給額児童1人 + 最小支給額児童2人 + (最小支給額児童3人目以降 * (対象児童人数 - 2))))

        return 手当条件 * 手当金額


class 児童扶養手当の全部支給所得条件(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "保護者への児童扶養手当の全額支給所得条件"
    reference = "https://www.cfa.go.jp/policies/hitori-oya/fuyou-teate/"
    documentation = """
    児童扶養手当制度
    """

    def formula(対象世帯, 対象期間, parameters):
        児童扶養手当 = parameters(対象期間).福祉.育児.児童扶養手当

        # 世帯で最も高い所得の人が基準となる
        世帯高所得 = 対象世帯("児童扶養手当の控除後世帯高所得", 対象期間)

        扶養人数 = 対象世帯("扶養人数", 対象期間)

        # 所得が全部支給所得制限限度額よりも高かったら一部支給になる
        # NOTE: 直接 `全部支給[扶養人数]` のように要素参照すると型が合わず複数世帯の場合に計算できないためnp.selectを使用
        全部支給所得制限限度額 = np.select(
            [扶養人数 == i for i in range(10)],
            [児童扶養手当.所得制限限度額.全部支給[i] for i in range(10)],
            -1).astype(int)

        全部支給所得条件 = 世帯高所得 < 全部支給所得制限限度額

        return 全部支給所得条件


class 児童扶養手当の一部支給所得条件(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "保護者への児童扶養手当の一部支給所得条件"
    reference = "https://www.cfa.go.jp/policies/hitori-oya/fuyou-teate/"
    documentation = """
    児童扶養手当制度
    """

    def formula(対象世帯, 対象期間, parameters):
        児童扶養手当 = parameters(対象期間).福祉.育児.児童扶養手当

        # 世帯で最も高い所得の人が基準となる
        世帯高所得 = 対象世帯("児童扶養手当の控除後世帯高所得", 対象期間)

        扶養人数 = 対象世帯("扶養人数", 対象期間)

        # NOTE: 直接 `全部支給[扶養人数]` のように要素参照すると型が合わず複数世帯の場合に計算できないためnp.selectを使用
        全部支給所得制限限度額 = np.select(
            [扶養人数 == i for i in range(10)],
            [児童扶養手当.所得制限限度額.全部支給[i] for i in range(10)],
            -1).astype(int)

        一部支給所得制限限度額 = np.select(
            [扶養人数 == i for i in range(10)],
            [児童扶養手当.所得制限限度額.一部支給[i] for i in range(10)],
            -1).astype(int)

        # 所得が全部支給所得制限限度額よりも高かったら一部支給になる
        # 所得が一部支給所得制限限度額よりも高かったら支給なしになる
        一部支給所得条件 = (世帯高所得 >= 全部支給所得制限限度額) * (世帯高所得 < 一部支給所得制限限度額)

        return 一部支給所得条件


class 児童扶養手当の対象児童人数(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "保護者への児童扶養手当の対象児童人数"
    reference = "https://www.cfa.go.jp/policies/hitori-oya/fuyou-teate/"
    documentation = """
    児童扶養手当制度
    """

    def formula(対象世帯, 対象期間, parameters):
        児童扶養手当 = parameters(対象期間).福祉.育児.児童扶養手当

        学年 = 対象世帯.members("学年", 対象期間)
        上限学年以下児童 = 学年 <= 児童扶養手当.上限学年

        # 特別児童扶養手当2級と同じ程度以上の障害を持つ児童は20歳未満まで対象
        # 参考：https://www.city.nagato.yamaguchi.jp/site/nagato-kosodatenavi/43285.html
        身体障害者手帳等級一覧 = 対象世帯.members("身体障害者手帳等級", 対象期間)
        療育手帳等級一覧 = 対象世帯.members("療育手帳等級", 対象期間)
        愛の手帳等級一覧 = 対象世帯.members("愛の手帳等級", 対象期間)
        精神障害者保健福祉手帳等級一覧 = 対象世帯.members("精神障害者保健福祉手帳等級", 対象期間)

        # 精神障害者保健福祉手帳の等級は以下の中標津町のHPを参照
        # https://www.nakashibetsu.jp/kurashi/kosodate_fukushi/shougaisha/teate/tokubetujidou/
        # 内部障害は対象になるか不明のため含めない
        対象障害者手帳等級 = \
            (身体障害者手帳等級一覧 == 身体障害者手帳等級パターン.一級) + \
            (身体障害者手帳等級一覧 == 身体障害者手帳等級パターン.二級) + \
            (身体障害者手帳等級一覧 == 身体障害者手帳等級パターン.三級) + \
            (療育手帳等級一覧 == 療育手帳等級パターン.A) + \
            (療育手帳等級一覧 == 療育手帳等級パターン.B) + \
            (愛の手帳等級一覧 == 愛の手帳等級パターン.一度) + \
            (愛の手帳等級一覧 == 愛の手帳等級パターン.二度) + \
            (愛の手帳等級一覧 == 愛の手帳等級パターン.三度) + \
            (精神障害者保健福祉手帳等級一覧 == 精神障害者保健福祉手帳等級パターン.一級) + \
            (精神障害者保健福祉手帳等級一覧 == 精神障害者保健福祉手帳等級パターン.二級)

        年齢 = 対象世帯.members("年齢", 対象期間)
        上限年齢未満障害児童 = 対象障害者手帳等級 * (年齢 < 児童扶養手当.障害児の上限年齢)
        対象児童人数 = 対象世帯.sum(上限学年以下児童 + 上限年齢未満障害児童)

        return 対象児童人数



"""
障害児童育成手当の実装
"""

import numpy as np
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯
from openfisca_japan.variables.障害.愛の手帳 import 愛の手帳等級パターン
from openfisca_japan.variables.障害.脳性まひ_進行性筋萎縮症 import 脳性まひ_進行性筋萎縮症パターン
from openfisca_japan.variables.障害.身体障害者手帳 import 身体障害者手帳等級パターン


class 障害児童育成手当(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "保護者への障害児童育成手当"
    reference = "https://www.city.shibuya.tokyo.jp/kodomo/ninshin/teate/jido_i.html"
    documentation = """
    渋谷区の児童育成（障害）手当

    - 〒150-8010 東京都渋谷区宇田川町1-1
    - 渋谷区子ども青少年課子育て給付係
    - 03-3463-2558
    """

    def formula(対象世帯, 対象期間, parameters):
        居住都道府県 = 対象世帯("居住都道府県", 対象期間)
        居住地条件 = 居住都道府県 == "東京都"

        障害児童育成手当 = parameters(対象期間).福祉.育児.障害児童育成手当

        # 世帯で最も高い所得の人が基準となる。特別児童扶養手当と同等の控除が適用される。
        # （参考）https://www.city.adachi.tokyo.jp/oyako/k-kyoiku/kosodate/hitorioya-ikuse.html
        世帯高所得 = 対象世帯("特別児童扶養手当の控除後世帯高所得", 対象期間)
        扶養人数 = 対象世帯("扶養人数", 対象期間)

        # NOTE: 直接 `所得制限限度額[扶養人数]` のように要素参照すると型が合わず複数世帯の場合に計算できないためnp.selectを使用
        所得制限限度額 = np.select(
            [扶養人数 == i for i in range(6)],
            [障害児童育成手当.所得制限限度額[i] for i in range(6)],
            -1).astype(int)

        所得条件 = 世帯高所得 < 所得制限限度額

        身体障害者手帳等級一覧 = 対象世帯.members("身体障害者手帳等級", 対象期間)
        愛の手帳等級一覧 = 対象世帯.members("愛の手帳等級", 対象期間)
        脳性まひ_進行性筋萎縮症一覧 = 対象世帯.members("脳性まひ_進行性筋萎縮症", 対象期間)
        年齢 = 対象世帯.members("年齢", 対象期間)
        児童である = 対象世帯.has_role(世帯.子)
        上限年齢未満の児童 = 児童である * (年齢 < 障害児童育成手当.上限年齢)

        対象障害者手帳等級 = \
            (身体障害者手帳等級一覧 == 身体障害者手帳等級パターン.一級) + \
            (身体障害者手帳等級一覧 == 身体障害者手帳等級パターン.二級) + \
            (愛の手帳等級一覧 == 愛の手帳等級パターン.一度) + \
            (愛の手帳等級一覧 == 愛の手帳等級パターン.二度) + \
            (愛の手帳等級一覧 == 愛の手帳等級パターン.三度) + \
            (脳性まひ_進行性筋萎縮症一覧 == 脳性まひ_進行性筋萎縮症パターン.有)

        上限年齢未満の身体障害を持つ児童人数 = 対象世帯.sum(上限年齢未満の児童 & 対象障害者手帳等級)
        手当金額 = 障害児童育成手当.金額 * 上限年齢未満の身体障害を持つ児童人数

        return 居住地条件 * 所得条件 * 手当金額



"""
傷病手当金の実装
"""

import numpy as np
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯, 人物
from openfisca_japan.variables.全般 import 就労形態パターン


class 傷病手当金_最大(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "傷病手当金"
    reference = "https://www.kyoukaikenpo.or.jp/g7/cat710/sb3160/sb3170/sbb31710/1950-271/"
    documentation = """
    休んだ日単位で支給されるため、算出額も一日当たりの金額とする。地方公務員の場合は一月当たりの日数が異なる。
    https://www.chikyosai.or.jp/division/short/scene/works/01.html

    算出方法は以下リンクも参考になる。
    https://www.asahi.com/relife/article/15231364
    """

    def formula(対象世帯, 対象期間, parameters):
        # 傷病手当金は連続3日間の休業のあと4日目から支給される（ただし労災による休業の場合を除く）
        けがによって連続三日以上休業している = 対象世帯.members("けがによって連続三日以上休業している", 対象期間)
        病気によって連続三日以上休業している = 対象世帯.members("病気によって連続三日以上休業している", 対象期間)

        業務によってけがをした = 対象世帯.members("業務によってけがをした", 対象期間)
        業務によって病気になった = 対象世帯.members("業務によって病気になった", 対象期間)

        休業条件 = (けがによって連続三日以上休業している * np.logical_not(業務によってけがをした)) +\
            (病気によって連続三日以上休業している * np.logical_not(業務によって病気になった))

        # 給与の支払いが無い場合対象
        休業中に給与の支払いがない = 対象世帯.members("休業中に給与の支払いがない", 対象期間)
        給与条件 = 休業中に給与の支払いがない

        # 傷病手当金は健康保険の加入者が対象
        # 75歳以上の場合後期高齢者医療制度の対象のため対象外
        # 参考: https://www.freee.co.jp/kb/kb-kaigyou/national_health_insurance/
        年齢 = 対象世帯.members("年齢", 対象期間)
        年齢条件 = 年齢 < 75

        一月当たりの日数_会社員 = parameters(対象期間).福祉.傷病手当金.日数_会社員
        一月当たりの日数_公務員 = parameters(対象期間).福祉.傷病手当金.日数_公務員

        # 自営業の場合国民健康保険に加入しているため対象外
        # 最大額のため、その他の場合も会社員とみなして計算
        就労形態 = 対象世帯.members("就労形態", 対象期間)
        就労形態ごとの倍率 = np.select(
            [就労形態 == 就労形態パターン.会社員,
             就労形態 == 就労形態パターン.公務員,
             就労形態 == 就労形態パターン.その他],
            [1 / 一月当たりの日数_会社員 * (2 / 3),
             1 / 一月当たりの日数_公務員 * (2 / 3),
             1 / 一月当たりの日数_会社員 * (2 / 3)],
            0)

        標準報酬月額 = 対象世帯.members("標準報酬月額", 対象期間)
        傷病手当金日額 = 標準報酬月額 * 就労形態ごとの倍率

        return 対象世帯.sum(休業条件 * 給与条件 * 年齢条件 * 傷病手当金日額)


class 傷病手当金_最小(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "傷病手当金"
    reference = "https://www.kyoukaikenpo.or.jp/g7/cat710/sb3160/sb3170/sbb31710/1950-271/"
    documentation = """
    休んだ日単位で支給されるため、算出額も一日当たりの金額とする。

    算出方法は以下リンクも参考になる。
    https://www.asahi.com/relife/article/15231364
    """

    def formula(対象世帯, 対象期間, _parameters):
        # NOTE: 以下を考慮し、最小額は0円とする
        # 任意継続被保険者である場合、傷病手当金の対象外
        # 労災保険から休業補償給付を受けている場合、傷病手当金の対象外
        # 会社の出産手当金、退職年金等を受け取っている場合、傷病手当金の対象外
        # 賞与は実際の標準報酬月額には含まれない（年2回以下の場合）ため、傷病手当金の最大額が実際よりも高く産出されてしまう
        return 0


class 休業中に給与の支払いがない(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "休業中に給与の支払いがない"
    documentation = """
    フロントエンドのフォームや制度にあわせて、支払いが「ない」方をTrueとする。
    """


class 業務によってけがをした(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "業務によってけがをしたか否か"


class 業務によって病気になった(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "業務によって病気になったか否か"


class 病気によって連続三日以上休業している(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "病気によって連続三日以上休業しているか否か"
    documentation = """
    TODO: 休業日数による判別が増えたら「休業日数」という変数にする
    """


class けがによって連続三日以上休業している(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "けがによって連続三日以上休業しているか否か"
    documentation = """
    TODO: 休業日数による判別が増えたら「休業日数」という変数にする
    """



"""
療育手帳の実装

（知的障害者を対象とする手帳。東京都などでは「愛の手帳」の名称だが等級が異なる）
"""

from openfisca_core.indexed_enums import Enum
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 人物


class 療育手帳等級パターン(Enum):
    __order__ = "無 A B"
    無 = "無"
    A = "A"
    B = "B"


class 療育手帳等級(Variable):
    value_type = Enum
    possible_values = 療育手帳等級パターン
    default_value = 療育手帳等級パターン.無
    entity = 人物
    definition_period = DAY
    label = "人物の療育手帳の等級"



"""
特別障害者手当の実装
"""

import numpy as np
from openfisca_core.periods import DAY, instant
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯, 人物
from openfisca_japan.variables.全般 import 高校生学年
from openfisca_japan.variables.障害.愛の手帳 import 愛の手帳等級パターン
from openfisca_japan.variables.障害.療育手帳 import 療育手帳等級パターン
from openfisca_japan.variables.障害.精神障害者保健福祉手帳 import 精神障害者保健福祉手帳等級パターン
from openfisca_japan.variables.障害.身体障害者手帳 import 身体障害者手帳等級パターン


class 障害基礎年金_最大(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "障害基礎年金の最大額"
    reference = "https://www.nenkin.go.jp/service/jukyu/shougainenkin/jukyu-yoken/20150514.html"
    documentation = """
    保険料納付済期間の条件（納付条件）については、利用者の回答負荷が高いため考慮せずに計算している。

    障害者手帳等級と障害基礎年金の等級には互換性がないため、最大値、最小額として幅を持たせて計算
    https://www.iizuka-nenkin.net/14097085633831
    """

    def formula(対象世帯, 対象期間, parameters):
        # NOTE: 幅を持たせて結果を算出しているため、回答負荷の高い初診日については考慮しない
        # (条件としては 初診日のときの年齢 < 65)

        # NOTE: 障害認定日の条件は回答負荷が高いため、過去の障害認定日と現在の障害が同様であると仮定して計算
        # ただし、障害認定日が20歳未満の場合20歳になってから支給が始まるため年齢条件のみ考慮
        年齢 = 対象世帯.members("年齢", 対象期間)
        年齢条件 = 年齢 >= 20

        支給額 = parameters(対象期間).福祉.障害基礎年金.支給額

        誕生年月日 = 対象世帯.members("誕生年月日", 対象期間)
        支給額変更日 = instant("1956-04-02").date
        一級の支給額 = np.select([誕生年月日 >= 支給額変更日],
            [支給額.支給額_一級_昭和三一年度以降],
            支給額.支給額_一級_昭和三〇年度以前)
        二級の支給額 = np.select([誕生年月日 >= 支給額変更日],
            [支給額.支給額_二級_昭和三一年度以降],
            支給額.支給額_二級_昭和三〇年度以前)

        障害等級 = 対象世帯.members("障害基礎年金_等級_最大", 対象期間)
        年金額 = np.select([障害等級 == 1, 障害等級 == 2],
            [一級の支給額, 二級の支給額],
            0)

        子の加算額 = 対象世帯("障害基礎年金の子の加算額_最大", 対象期間)

        # NOTE: 年金額が0円出ない場合のみ加算額を足したいため、障害等級を改めて条件に指定
        障害条件 = (障害等級 == 1) + (障害等級 == 2)
        支給条件 = 年齢条件 * 障害条件

        # 月額に直すため12か月で割る
        # NOTE: 子の加算額は世帯単位のため、支給条件を満たす世帯員が1人でもいる場合は加算する
        return (対象世帯.sum(支給条件 * 年金額) + 対象世帯.any(支給条件) * 子の加算額) / 12


class 障害基礎年金_最小(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "障害基礎年金の最小額"
    reference = "https://www.nenkin.go.jp/service/jukyu/shougainenkin/jukyu-yoken/20150514.html"
    documentation = """
    保険料納付済期間の条件（納付条件）については、利用者の回答負荷が高いため考慮せずに計算している。

    障害者手帳等級と障害基礎年金の等級には互換性がないため、最大値、最小額として幅を持たせて計算
    https://www.iizuka-nenkin.net/14097085633831
    """

    def formula(対象世帯, 対象期間, parameters):
        # NOTE: 幅を持たせて結果を算出しているため、回答負荷の高い初診日については考慮しない
        # (条件としては 初診日のときの年齢 < 65)

        # NOTE: 障害認定日の条件は回答負荷が高いため、過去の障害認定日と現在の障害が同様であると仮定して計算
        # ただし、障害認定日が20歳未満の場合20歳になってから支給が始まるため年齢条件のみ考慮
        年齢 = 対象世帯.members("年齢", 対象期間)
        年齢条件 = 年齢 >= 20

        支給額 = parameters(対象期間).福祉.障害基礎年金.支給額

        誕生年月日 = 対象世帯.members("誕生年月日", 対象期間)
        支給額変更日 = instant("1956-04-02").date
        一級の支給額 = np.select([誕生年月日 >= 支給額変更日],
            [支給額.支給額_一級_昭和三一年度以降],
            支給額.支給額_一級_昭和三〇年度以前)
        二級の支給額 = np.select([誕生年月日 >= 支給額変更日],
            [支給額.支給額_二級_昭和三一年度以降],
            支給額.支給額_二級_昭和三〇年度以前)

        障害等級 = 対象世帯.members("障害基礎年金_等級_最小", 対象期間)
        年金額 = np.select([障害等級 == 1, 障害等級 == 2],
            [一級の支給額, 二級の支給額],
            0)

        子の加算額 = 対象世帯("障害基礎年金の子の加算額_最小", 対象期間)

        # NOTE: 年金額が0円出ない場合のみ加算額を足したいため、障害等級を改めて条件に指定
        障害条件 = (障害等級 == 1) + (障害等級 == 2)
        支給条件 = 年齢条件 * 障害条件

        # 月額に直すため12か月で割る
        # NOTE: 子の加算額は世帯単位のため、支給条件を満たす世帯員が1人でもいる場合は加算する
        return (対象世帯.sum(支給条件 * 年金額) + 対象世帯.any(支給条件) * 子の加算額) / 12


class 障害基礎年金_等級_最大(Variable):
    value_type = int
    entity = 人物
    definition_period = DAY
    label = "障害基礎年金の障害等級の最大値"
    reference = "https://www.nenkin.go.jp/service/jukyu/shougainenkin/jukyu-yoken/20150514.html"
    documentation = """
    障害者手帳等級と障害基礎年金の等級には互換性がないため、手帳等級から年金の等級を幅を持たせて推定
    https://www.iizuka-nenkin.net/14097085633831

    すでに受給していないと等級は分からないため、直接利用者が入力する形式にはしない
    該当しない場合は0とする
    """

    def formula(対象人物, 対象期間, _parameters):
        身体障害者手帳等級 = 対象人物("身体障害者手帳等級", 対象期間)
        精神障害者保健福祉手帳等級 = 対象人物("精神障害者保健福祉手帳等級", 対象期間)
        愛の手帳等級 = 対象人物("愛の手帳等級", 対象期間)
        療育手帳等級 = 対象人物("療育手帳等級", 対象期間)

        # 障害等級の条件（障害の程度2級以上）
        # NOTE: 障害者手帳等級と障害基礎年金の等級には互換性がないため、該当しうる手帳等級の最小値を用いて計算

        # 身体障害者手帳等級4級「音声又は言語機能に著しい障害を有するもの」→障害基礎年金2級
        # 身体障害者手帳等級3級「両下肢を足関節以上で欠くもの」→ 障害基礎年金1級
        # 障害基礎年金: https://www.nenkin.go.jp/service/jukyu/shougainenkin/ninteikijun/tokyuhyo.html
        # 身体障害者手帳: https://www.jsh-japan.jp/cordiale-farm/column/3122/#techo2
        一級の身体障害条件 = (身体障害者手帳等級 == 身体障害者手帳等級パターン.一級) +\
            (身体障害者手帳等級 == 身体障害者手帳等級パターン.二級) +\
            (身体障害者手帳等級 == 身体障害者手帳等級パターン.三級)
        二級の身体障害条件 = (身体障害者手帳等級 == 身体障害者手帳等級パターン.四級)

        # 精神障害者手帳の1~2級: 障害年金とほぼ同じ基準
        # https://www.iizuka-nenkin.net/14097085633831
        一級の精神障害条件 = 精神障害者保健福祉手帳等級 == 精神障害者保健福祉手帳等級パターン.一級
        二級の精神障害条件 = 精神障害者保健福祉手帳等級 == 精神障害者保健福祉手帳等級パターン.二級

        # 療育手帳（愛の手帳）中度: 障害基礎年金1級~2級の可能性を検討
        # https://www.nenkin.go.jp/service/jukyu/shougainenkin/ninteikijun/20160715.files/A.pdf
        # https://hyogo-osaka-shogai.com/wp/chiteki-2
        一級の知的障害条件 = (愛の手帳等級 == 愛の手帳等級パターン.一度) +\
            (愛の手帳等級 == 愛の手帳等級パターン.二度) +\
            (愛の手帳等級 == 愛の手帳等級パターン.三度) +\
            (療育手帳等級 == 療育手帳等級パターン.A) +\
            (療育手帳等級 == 療育手帳等級パターン.B)

        障害等級 = np.select([
            一級の身体障害条件 + 一級の精神障害条件 + 一級の知的障害条件,
            二級の身体障害条件 + 二級の精神障害条件],
            [1, 2],
            0)
        return 障害等級


class 障害基礎年金_等級_最小(Variable):
    value_type = int
    entity = 人物
    definition_period = DAY
    label = "障害基礎年金の障害等級の最小値"
    reference = "https://www.nenkin.go.jp/service/jukyu/shougainenkin/jukyu-yoken/20150514.html"
    documentation = """
    障害者手帳等級と障害基礎年金の等級には互換性がないため、手帳等級から年金の等級を幅を持たせて推定
    https://www.iizuka-nenkin.net/14097085633831

    すでに受給していないと等級は分からないため、直接利用者が入力する形式にはしない
    該当しない場合は0とする
    """

    def formula(対象人物, 対象期間, _parameters):
        # 障害等級の条件（障害の程度2級以上）
        # NOTE: 障害者手帳等級と障害基礎年金の等級には互換性がないため、比較的対応がとれている精神障害者手帳以外は最小額0円とする
        精神障害者保健福祉手帳等級 = 対象人物("精神障害者保健福祉手帳等級", 対象期間)

        # 精神障害者手帳の1~2級: 障害年金とほぼ同じ基準
        # https://www.iizuka-nenkin.net/14097085633831
        一級の精神障害条件 = 精神障害者保健福祉手帳等級 == 精神障害者保健福祉手帳等級パターン.一級
        二級の精神障害条件 = 精神障害者保健福祉手帳等級 == 精神障害者保健福祉手帳等級パターン.二級

        障害等級 = np.select([一級の精神障害条件, 二級の精神障害条件],
            [1, 2],
            0)
        return 障害等級


class 障害基礎年金の子の加算額_最大(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "障害基礎年金の子の加算額の最大額"
    reference = "https://www.nenkin.go.jp/service/jukyu/shougainenkin/jukyu-yoken/20150514.html"

    def formula(対象世帯, 対象期間, parameters):
        対象児童人数 = 対象世帯("障害基礎年金の対象児童人数_最大", 対象期間)

        支給額 = parameters(対象期間).福祉.障害基礎年金.支給額

        # 複数世帯入力(2以上の長さのndarray入力)対応のためndarray化
        return np.select([対象児童人数 <= 2, 対象児童人数 > 2],
            [支給額.子の加算額_二人目まで * 対象児童人数,
             支給額.子の加算額_二人目まで * 2 + 支給額.子の加算額_三人目以降 * (対象児童人数 - 2)],
            0)


class 障害基礎年金の子の加算額_最小(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "障害基礎年金の子の加算額の最小額"
    reference = "https://www.nenkin.go.jp/service/jukyu/shougainenkin/jukyu-yoken/20150514.html"

    def formula(対象世帯, 対象期間, parameters):
        対象児童人数 = 対象世帯("障害基礎年金の対象児童人数_最小", 対象期間)

        支給額 = parameters(対象期間).福祉.障害基礎年金.支給額

        # 複数世帯入力(2以上の長さのndarray入力)対応のためndarray化
        return np.select([対象児童人数 <= 2, 対象児童人数 > 2],
            [支給額.子の加算額_二人目まで * 対象児童人数,
             支給額.子の加算額_二人目まで * 2 + 支給額.子の加算額_三人目以降 * (対象児童人数 - 2)],
            0)


class 障害基礎年金の対象児童人数_最大(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "障害基礎年金の対象児童人数の最大値"
    reference = "https://www.nenkin.go.jp/service/jukyu/shougainenkin/jukyu-yoken/20150514.html"
    documentation = """
    障害等級の算出に幅があるため、対象人数も最大、最小で算出
    """

    def formula(対象世帯, 対象期間, _parameters):
        障害等級 = 対象世帯.members("障害基礎年金_等級_最大", 対象期間)
        子である = 対象世帯.has_role(世帯.子)
        年齢 = 対象世帯.members("年齢", 対象期間)
        二十歳未満で障害等級二級以上の子 = 子である * (年齢 < 20) * ((障害等級 == 1) + (障害等級 == 2))

        # NOTE: 入力情報では誕生年月日が分からないため学年を使った計算で代替
        # （18歳になった後の最初の3月31日までの子）
        学年 = 対象世帯.members("学年", 対象期間)
        十八歳以下の子 = 子である * (学年 <= 高校生学年.三年生.value)

        return 対象世帯.sum(十八歳以下の子 + 二十歳未満で障害等級二級以上の子)


class 障害基礎年金の対象児童人数_最小(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "障害基礎年金の対象児童人数の最小値"
    reference = "https://www.nenkin.go.jp/service/jukyu/shougainenkin/jukyu-yoken/20150514.html"
    documentation = """
    障害等級の算出に幅があるため、対象人数も最大、最小で算出
    """

    def formula(対象世帯, 対象期間, _parameters):
        障害等級 = 対象世帯.members("障害基礎年金_等級_最小", 対象期間)
        子である = 対象世帯.has_role(世帯.子)
        年齢 = 対象世帯.members("年齢", 対象期間)
        # 障害等級は数字が小さいほど等級が高い
        二十歳未満で障害等級二級以上の子 = 子である * (年齢 < 20) * ((障害等級 == 1) + (障害等級 == 2))

        # NOTE: 入力情報では誕生年月日が分からないため学年を使った計算で代替
        # （18歳になった後の最初の3月31日までの子）
        学年 = 対象世帯.members("学年", 対象期間)
        十八歳以下の子 = 子である * (学年 <= 高校生学年.三年生.value)

        return 対象世帯.sum(十八歳以下の子 + 二十歳未満で障害等級二級以上の子)



"""
預貯金の実装
"""

from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯, 人物


class 預貯金(Variable):
    value_type = int
    default_value = 0
    entity = 人物
    definition_period = DAY
    label = "預貯金"
    reference = "https://www.fukushi.metro.tokyo.lg.jp/seikatsu/teisyotokusyataisaku/jukenseichallenge.html"


class 子供の預貯金(Variable):
    value_type = int
    default_value = 0
    entity = 世帯
    definition_period = DAY
    label = "子供の預貯金"
    reference = "https://www.fukushi.metro.tokyo.lg.jp/seikatsu/teisyotokusyataisaku/jukenseichallenge.html"

    def formula(対象世帯, 対象期間, parameters):
        if 対象世帯.nb_persons(世帯.子) == 0:
            return 0

        子供である = 対象世帯.has_role(世帯.子)
        預貯金一覧 = 対象世帯.members("預貯金", 対象期間)
        return 対象世帯.sum(子供である * 預貯金一覧)



"""
被災者生活再建支援制度の実装
"""


from functools import cache

import numpy as np
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan import COUNTRY_DIR
from openfisca_japan.entities import 世帯
from openfisca_japan.variables.災害.住宅 import 住宅再建方法パターン, 住宅被害パターン


@cache
def 基礎支援金額表():
    """
    csvファイルから値を読み込み

    基礎支援金額表()[住宅被害] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/災害/支援/被災者生活再建支援制度_基礎支援金.csv",
                  delimiter=",", skip_header=1, dtype="int64")[1:]


@cache
def 加算支援金額表():
    """
    csvファイルから値を読み込み

    加算支援金額表()[住宅被害, 住宅再建方法] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/災害/支援/被災者生活再建支援制度_加算支援金.csv",
                  delimiter=",", skip_header=1, dtype="int64")[:, 1:]


class 被災者生活再建支援法の適用地域である(Variable):
    value_type = bool
    entity = 世帯
    definition_period = DAY
    label = "被災者生活再建支援法の適用地域であるかどうか"
    reference = "https://www.bousai.go.jp/taisaku/seikatsusaiken/pdf/140612gaiyou.pdf"
    documentation = """
    自然災害ごとに、市区町村ごとに適用有無が決まる
    災害発生時は適用有無が変化していくので注意
    """
    # TODO: 居住市区町村から支援法が適用されているかどうかを計算できるようにする


class 被災者生活再建支援制度(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "被災者生活再建支援制度"
    reference = "https://www.bousai.go.jp/taisaku/seikatsusaiken/pdf/140612gaiyou.pdf"
    documentation = """
    対象となる自然災害は都道府県が公示

    算出方法は以下リンクも参考になる。
    https://www.bousai.go.jp/taisaku/hisaisyagyousei/pdf/kakusyuseido_tsuujou.pdf
    """

    def formula(対象世帯, 対象期間, _parameters):
        被災者生活再建支援法の適用地域である = 対象世帯("被災者生活再建支援法の適用地域である", 対象期間)
        基礎支援金 = 対象世帯("被災者生活再建支援制度_基礎支援金", 対象期間)
        加算支援金 = 対象世帯("被災者生活再建支援制度_加算支援金", 対象期間)

        世帯人数 = 対象世帯("世帯人数", 対象期間)
        世帯人数に応じた倍率 = np.select(
            [世帯人数 == 1],
            [0.75],
            1)

        return 被災者生活再建支援法の適用地域である * 世帯人数に応じた倍率 * (基礎支援金 + 加算支援金)


class 被災者生活再建支援制度_基礎支援金(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "被災者生活再建支援制度における基礎支援金"
    reference = "https://www.bousai.go.jp/taisaku/seikatsusaiken/pdf/140612gaiyou.pdf"

    def formula(対象世帯, 対象期間, _parameters):
        住宅被害 = 対象世帯("住宅被害", 対象期間)
        支援制度対象である = 住宅被害 != 住宅被害パターン.無

        住宅被害区分 = np.select(
            [住宅被害 == 住宅被害パターン.滅失または流失,
             住宅被害 == 住宅被害パターン.全壊,
             住宅被害 == 住宅被害パターン.解体,
             住宅被害 == 住宅被害パターン.長期避難,
             住宅被害 == 住宅被害パターン.大規模半壊,
             住宅被害 == 住宅被害パターン.中規模半壊],
            list(range(6)),
            -1).astype(int)  # intにできるようデフォルトをNoneではなく-1

        基礎支援金額 = 基礎支援金額表()[住宅被害区分]
        return 支援制度対象である * 基礎支援金額


class 被災者生活再建支援制度_加算支援金(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "被災者生活再建支援制度における加算支援金"
    reference = "https://www.bousai.go.jp/taisaku/seikatsusaiken/pdf/140612gaiyou.pdf"

    def formula(対象世帯, 対象期間, _parameters):
        住宅被害 = 対象世帯("住宅被害", 対象期間)
        住宅再建方法 = 対象世帯("住宅再建方法", 対象期間)
        支援制度対象である = (住宅被害 != 住宅被害パターン.無) * (住宅再建方法 != 住宅再建方法パターン.無)

        住宅被害区分 = np.select(
            [住宅被害 == 住宅被害パターン.滅失または流失,
             住宅被害 == 住宅被害パターン.全壊,
             住宅被害 == 住宅被害パターン.解体,
             住宅被害 == 住宅被害パターン.長期避難,
             住宅被害 == 住宅被害パターン.大規模半壊,
             住宅被害 == 住宅被害パターン.中規模半壊],
            list(range(6)),
            -1).astype(int)  # intにできるようデフォルトをNoneではなく-1

        住宅再建方法区分 = np.select(
            [住宅再建方法 == 住宅再建方法パターン.建設または購入,
             住宅再建方法 == 住宅再建方法パターン.補修,
             住宅再建方法 == 住宅再建方法パターン.賃借],
            list(range(3)),
            -1).astype(int)  # intにできるようデフォルトをNoneではなく-1

        加算支援金額 = 加算支援金額表()[住宅被害区分, 住宅再建方法区分]
        return 支援制度対象である * 加算支援金額



"""
標準報酬月額の実装
"""

import numpy as np
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 人物


class 標準報酬月額(Variable):
    value_type = float
    entity = 人物
    definition_period = DAY
    label = "標準報酬月額"
    reference = "https://www.kyoukaikenpo.or.jp/g3/cat320/sb3160/sbb3165/1962-231/"
    documentation = """
    被保険者が事業主から受ける毎月の報酬

    基本給のほか、役付手当、勤務地手当、家族手当、通勤手当、住宅手当、残業手当等も含まれる
    賞与は年3回以下である場合含まれない。
    また、厳密には毎月の報酬額を一定区分ごとに等級付けするため実際の収入より増減する場合がある。
    """

    def formula(対象人物, 対象期間, parameters):
        収入 = 対象人物("収入", 対象期間)

        # TODO: 賞与を除いて計算する（賞与を別フォームで入力するようになったら対応）
        # 現状では標準報酬月額が高めに算出される

        # 等級ごとに段階的に標準報酬月額が定められているため、金額に上限が存在する
        最大額 = parameters(対象期間).標準報酬月額.標準報酬月額_最大等級

        return np.clip(収入 / 12, None, 最大額)



"""
児童育成手当の実装
"""

import numpy as np
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯


class 児童育成手当(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "保護者への児童手当"
    reference = "https://www.city.shibuya.tokyo.jp/kodomo/teate/hitorioya/hitorioya_teate.html"
    documentation = """
    渋谷区の児童育成手当制度

    - 〒150-8010 東京都渋谷区宇田川町1-1
    - 渋谷区子ども青少年課子育て給付係
    - 03-3463-2558
    """

    def formula(対象世帯, 対象期間, parameters):
        居住都道府県 = 対象世帯("居住都道府県", 対象期間)
        居住地条件 = 居住都道府県 == "東京都"

        児童育成手当 = parameters(対象期間).福祉.育児.児童育成手当

        # 世帯で最も高い所得の人が基準となる。特別児童扶養手当と同等の控除が適用される。
        # （参考）https://www.city.adachi.tokyo.jp/oyako/k-kyoiku/kosodate/hitorioya-ikuse.html
        世帯高所得 = 対象世帯("特別児童扶養手当の控除後世帯高所得", 対象期間)
        扶養人数 = 対象世帯("扶養人数", 対象期間)

        # NOTE: 直接 `所得制限限度額[扶養人数]` のように要素参照すると型が合わず複数世帯の場合に計算できないためnp.selectを使用
        所得制限限度額 = np.select(
            [扶養人数 == i for i in range(6)],
            [児童育成手当.所得制限限度額[i] for i in range(6)],
            -1).astype(int)

        所得条件 = 世帯高所得 < 所得制限限度額

        ひとり親世帯である = 対象世帯("ひとり親", 対象期間)
        学年 = 対象世帯.members("学年", 対象期間)
        上限学年以下の人数 = 対象世帯.sum(学年 <= 児童育成手当.上限学年)
        手当条件 = ひとり親世帯である * 所得条件 * 居住地条件

        return 手当条件 * 上限学年以下の人数 * 児童育成手当.金額



"""
高等学校等就学支援金の実装
"""

from functools import cache

import numpy as np
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan import COUNTRY_DIR
from openfisca_japan.entities import 世帯
from openfisca_japan.variables.全般 import 高校生学年
from openfisca_japan.variables.福祉.育児.高等学校奨学給付金 import 高校履修種別パターン, 高校運営種別パターン

# TODO: 専攻科の就学支援金についても実装する（高等学校等就学支援金制度では専攻科は対象外）

# NOTE: 項目数が多い金額表は可読性の高いCSV形式としている。


@cache
def 支給限度額_学年制表():
    """
    csvファイルから値を取得

    支給限度額_学年制表()[高校履修種別, 高校運営種別] の形で参照可能
    """
    # NOTE: 特別支援学校等、一部の高校履修種別は非対応（網羅すると判別のために利用者の入力負担が増えてしまうため）
    # https://www.mext.go.jp/a_menu/shotou/mushouka/__icsFiles/afieldfile/2020/04/30/100014428_4.pdf
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/育児/高等学校等就学支援金/支給額/支給限度額_学年制.csv",
                         delimiter=",", skip_header=1, dtype="int64")[:, 1:]


@cache
def 支給限度額_単位制表():
    """
    csvファイルから値を取得

    支給限度額_単位制表()[高校履修種別, 高校運営種別] の形で参照可能
    """
    # 月額の最大値として、年間取得可能最大単位数を取った場合の年額を12か月で按分した値を使用
    # https://www.mext.go.jp/a_menu/shotou/mushouka/__icsFiles/afieldfile/2020/04/30/100014428_4.pdf
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/育児/高等学校等就学支援金/支給額/支給限度額_単位制.csv",
                         delimiter=",", skip_header=1, dtype="int64")[:, 1:]


@cache
def 加算額_学年制表():
    """
    csvファイルから値を取得

    加算額_学年制表()[高校履修種別, 高校運営種別] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/育児/高等学校等就学支援金/支給額/加算額_学年制.csv",
                         delimiter=",", skip_header=1, dtype="int64")[:, 1:]


@cache
def 加算額_単位制表():
    """
    csvファイルから値を取得

    加算額_単位制表()[高校履修種別][高校運営種別] の形で参照可能
    """
    # 月額の最大値として、年間取得可能最大単位数を取った場合の年額を12か月で按分した値を使用
    # https://www.mext.go.jp/a_menu/shotou/mushouka/__icsFiles/afieldfile/2020/04/30/100014428_4.pdf
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/育児/高等学校等就学支援金/支給額/加算額_単位制.csv",
                         delimiter=",", skip_header=1, dtype="int64")[:, 1:]


class 高等学校等就学支援金_最小(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "高等学校等就学支援金"
    reference = "https://www.mext.go.jp/a_menu/shotou/mushouka/1342674.htm"
    documentation = """
    専攻科は対象外。
    算出方法は以下リンクも参考になる。
    (条件) https://www.mext.go.jp/a_menu/shotou/mushouka/20220329-mxt_kouhou02-3.pdf
    (金額) https://www.mext.go.jp/a_menu/shotou/mushouka/__icsFiles/afieldfile/2020/04/30/100014428_4.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        所得判定基準 = 対象世帯("高等学校等就学支援金_所得判定基準", 対象期間)
        学年 = 対象世帯.members("学年", 対象期間)
        高校生である = (学年 >= 高校生学年.一年生.value) * (学年 <= 高校生学年.三年生.value)

        高校履修種別 = 対象世帯.members("高校履修種別", 対象期間)
        高校運営種別 = 対象世帯.members("高校運営種別", 対象期間)

        高校履修種別インデックス = np.select(
            [高校履修種別 == 高校履修種別パターン.全日制課程,
             高校履修種別 == 高校履修種別パターン.定時制課程,
             高校履修種別 == 高校履修種別パターン.通信制課程],
            list(range(3)),
            -1).astype(int)

        高校運営種別インデックス = np.select(
            [高校運営種別 == 高校運営種別パターン.公立,
             高校運営種別 == 高校運営種別パターン.国立,
             高校運営種別 == 高校運営種別パターン.私立],
            list(range(3)),
            -1).astype(int)

        支給対象者である = 高校生である * (高校履修種別 != 高校履修種別パターン.無) * (高校運営種別 != 高校運営種別パターン.無)

        支給金額 = 支給限度額_学年制表()[高校履修種別インデックス, 高校運営種別インデックス]
        加算金額 = 加算額_学年制表()[高校履修種別インデックス, 高校運営種別インデックス]

        合計支給金額 = 対象世帯.sum(支給対象者である * 支給金額)
        合計加算金額 = 対象世帯.sum(支給対象者である * 加算金額)

        所得が支給対象である = 所得判定基準 < parameters(対象期間).福祉.育児.高等学校等就学支援金.所得判定基準.所得判定基準
        所得が加算対象である = 所得判定基準 < parameters(対象期間).福祉.育児.高等学校等就学支援金.所得判定基準.加算_所得判定基準

        return 合計支給金額 * 所得が支給対象である + 合計加算金額 * 所得が加算対象である


class 高等学校等就学支援金_最大(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "高等学校等就学支援金"
    reference = "https://www.mext.go.jp/a_menu/shotou/mushouka/1342674.htm"
    documentation = """
    専攻科は対象外。
    算出方法は以下リンクも参考になる。
    (条件) https://www.mext.go.jp/a_menu/shotou/mushouka/20220329-mxt_kouhou02-3.pdf
    (金額) https://www.mext.go.jp/a_menu/shotou/mushouka/__icsFiles/afieldfile/2020/04/30/100014428_4.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        所得判定基準 = 対象世帯("高等学校等就学支援金_所得判定基準", 対象期間)
        学年 = 対象世帯.members("学年", 対象期間)
        高校生である = (学年 >= 高校生学年.一年生.value) * (学年 <= 高校生学年.三年生.value)

        高校履修種別 = 対象世帯.members("高校履修種別", 対象期間)
        高校運営種別 = 対象世帯.members("高校運営種別", 対象期間)

        高校履修種別インデックス = np.select(
            [高校履修種別 == 高校履修種別パターン.全日制課程,
             高校履修種別 == 高校履修種別パターン.定時制課程,
             高校履修種別 == 高校履修種別パターン.通信制課程],
            list(range(3)),
            -1).astype(int)

        高校運営種別インデックス = np.select(
            [高校運営種別 == 高校運営種別パターン.公立,
             高校運営種別 == 高校運営種別パターン.国立,
             高校運営種別 == 高校運営種別パターン.私立],
            list(range(3)),
            -1).astype(int)

        支給対象者である = 高校生である * (高校履修種別 != 高校履修種別パターン.無) * (高校運営種別 != 高校運営種別パターン.無)

        支給金額 = 支給限度額_単位制表()[高校履修種別インデックス, 高校運営種別インデックス]
        加算金額 = 加算額_単位制表()[高校履修種別インデックス, 高校運営種別インデックス]

        合計支給金額 = 対象世帯.sum(支給対象者である * 支給金額)
        合計加算金額 = 対象世帯.sum(支給対象者である * 加算金額)

        所得が支給対象である = 所得判定基準 < parameters(対象期間).福祉.育児.高等学校等就学支援金.所得判定基準.所得判定基準
        所得が加算対象である = 所得判定基準 < parameters(対象期間).福祉.育児.高等学校等就学支援金.所得判定基準.加算_所得判定基準

        return 合計支給金額 * 所得が支給対象である + 合計加算金額 * 所得が加算対象である


class 高等学校等就学支援金_所得判定基準(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "高等学校等就学支援金計算において、支給額の判定基準となる額"
    reference = "https://www.mext.go.jp/a_menu/shotou/mushouka/20220329-mxt_kouhou02-3.pdf"

    def formula(対象世帯, 対象期間, parameters):
        課税標準額 = 対象世帯("控除後住民税世帯高所得", 対象期間)
        調整控除 = 対象世帯("調整控除", 対象期間)

        二人親である = 対象世帯.nb_persons(世帯.親) == 2
        課税標準額 = 課税標準額 + 二人親である * 対象世帯("世帯主の配偶者の控除後住民税所得", 対象期間)
        調整控除 = 調整控除 + 二人親である * 対象世帯("世帯主の配偶者の調整控除", 対象期間)

        return 課税標準額 * 0.06 - 調整控除


class 世帯主の配偶者の控除後住民税所得(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "世帯主の配偶者の控除後住民税所得"
    reference = "https://www.town.hinode.tokyo.jp/0000000516.html"
    # TODO: 各種控除のentityが人物になったら削除
    # （世帯主が「自分」または「配偶者」でないと実際の控除額がずれるため）

    def formula(対象世帯, 対象期間, parameters):
        # 所得が高いほうが控除を受ける対象となる
        所得一覧 = 対象世帯.members("所得", 対象期間)
        所得降順 = 対象世帯.get_rank(対象世帯, - 所得一覧, condition=対象世帯.has_role(世帯.親))
        納税者の配偶者である = 所得降順 == 1
        納税者の配偶者の所得 = 対象世帯.sum(所得一覧 * 納税者の配偶者である)

        # 扶養等と関係のない、納税者全員に関わる控除を追加
        # TODO: 世帯高所得ではなく自分、配偶者それぞれの所得から控除額を算出
        基礎控除 = np.select(
            [納税者の配偶者の所得 <= 24000000,
             (納税者の配偶者の所得 > 24000000) * (納税者の配偶者の所得 <= 24500000),
             (納税者の配偶者の所得 > 24500000) * (納税者の配偶者の所得 <= 25000000)],
            [430000,
             290000,
             150000],
            0)

        # 他の控除（雑損控除・医療費控除等）は定額でなく実費を元に算出するため除外する
        総控除額 = 基礎控除

        # 負の数にならないよう、0円未満になった場合は0円に補正
        return np.clip(納税者の配偶者の所得 - 総控除額, 0.0, None)


class 世帯主の配偶者の人的控除額の差(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "世帯主の配偶者に対する所得税と住民税の控除の差額"
    reference = "https://www.town.hinode.tokyo.jp/0000000516.html"
    # TODO: 各種控除のentityが人物になったら削除
    # （世帯主が「自分」または「配偶者」でないと実際の額とずれるため）

    def formula(対象世帯, 対象期間, parameters):
        # 所得が高いほうが控除を受ける対象となる
        所得一覧 = 対象世帯.members("所得", 対象期間)
        所得降順 = 対象世帯.get_rank(対象世帯, - 所得一覧, condition=対象世帯.has_role(世帯.親))
        納税者の配偶者である = 所得降順 == 1
        納税者の配偶者の所得 = 対象世帯.sum(所得一覧 * 納税者の配偶者である)

        基礎控除差額 = np.where(納税者の配偶者の所得 <= 25000000, 50000, 0)

        return 基礎控除差額


class 世帯主の配偶者の調整控除(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "世帯主の配偶者の調整控除"
    reference = "https://money-bu-jpx.com/news/article043882/"
    # TODO: 各種控除のentityが人物になったら削除
    # （世帯主が「自分」または「配偶者」でないと実際の額とずれるため）

    def formula(対象世帯, 対象期間, _parameters):
        人的控除額の差 = 対象世帯("世帯主の配偶者の人的控除額の差", 対象期間)

        # 個人住民税の課税所得金額に相当
        控除後住民税世帯高所得 = 対象世帯("世帯主の配偶者の控除後住民税所得", 対象期間)

        控除額 = np.select(
            [控除後住民税世帯高所得 <= 2000000,
             (控除後住民税世帯高所得 > 2000000) * (控除後住民税世帯高所得 < 25000000)],
            # (noqa) np.maxを世帯員抽出ではなく2つの式の比較に使用しているためlinterを許容
            [np.min([控除後住民税世帯高所得, 人的控除額の差], axis=0) * 0.05,  # noqa: TID251
             (人的控除額の差 - (控除後住民税世帯高所得 - 2000000)) * 0.05],
            0)

        # 負の数にならないよう、0円未満になった場合は0円に補正
        return np.clip(控除額, 0.0, None)
